import ast
import inspect
import os
import re
import subprocess
import sys

from .. import install

__all__ = ["autoinstall"]


def _get_stdlib_module_names():
    if sys.version_info >= (3, 10):
        return sys.stdlib_module_names
    else:
        # The following import causes warnings in python >= 3.10
        import distutils.sysconfig as sysconfig

        stdlib_module_names = set()
        std_lib = sysconfig.get_python_lib(standard_lib=True)

        for top, dirs, files in os.walk(std_lib):
            for nm in files:
                prefix = top[len(std_lib) + 1 :]
                if prefix[:13] == "site-packages":
                    continue
                if nm == "__init__.py":
                    stdlib_module_names.add(
                        top[len(std_lib) + 1 :].replace(os.path.sep, ".")
                    )
                elif nm[-3:] == ".py":
                    stdlib_module_names.add(
                        os.path.join(prefix, nm)[:-3].replace(os.path.sep, ".")
                    )
                elif nm[-3:] == ".so" and top[-11:] == "lib-dynload":
                    stdlib_module_names.add(nm[0:-3])
        for builtin in sys.builtin_module_names:
            stdlib_module_names.add(builtin)
        return stdlib_module_names


def _get_deps(path, include_stdlib=False):
    stdlib_module_names = _get_stdlib_module_names()
    with open(path) as f:
        root = ast.parse(f.read(), path)

    deps = []
    for node in ast.walk(root):
        if isinstance(node, ast.Import):
            for n in node.names:
                mod = n.name.split(".")[0]
                if include_stdlib or mod not in stdlib_module_names:
                    deps.append(mod)
        elif isinstance(node, ast.ImportFrom):
            mod = node.module.split(".")[0]
            if include_stdlib or mod not in stdlib_module_names:
                deps.append(mod)
        else:
            continue

    # allow comments to specify package name, e.g....
    # import sklearn  # install scikit-learn
    # TODO Make this robust a la https://stackoverflow.com/a/36055400/2903532
    with open(path) as f:
        install_lines = [
            line
            for line in f
            if "import" in line and re.search(r"#\s*(?:skip\s+)?install\s+", line)
        ]
    skips = {
        dep
        for dep in deps
        for line in install_lines
        if f" {dep}" in line and re.search(r"#\s*skip\s+install", line)
    }
    dep_pkg_list = []
    for dep in deps:
        if dep in skips:
            continue
        pkgs = [
            m.group(1).split()
            for line in install_lines
            if f" {dep}" in line
            and (m := re.search(r"#\s+install\s+([\s0-9A-Za-z-_.]+)", line))
        ]
        pkgs = set(pkg for group in pkgs for pkg in group if pkg)  # flatten pkgs
        dep_pkg_list.append((dep, tuple(pkgs) or None))
    return dep_pkg_list


def _is_installed(pkg):
    """Check whether pkg is already installed."""
    result = subprocess.run([sys.executable, "-m", "pip", "show", pkg])
    return not result.returncode


def autoinstall(interactive=True, **kwargs):
    """Install all packages imported in the module where this function is
    called.  Packages already installed will be skipped unless upgrade=True.

    If the package/distribution name to be installed is different from the
    module name (e.g. scikit-learn vs sklearn), the package name(s) should be
    provided in an in-line comment beginning with install:
    `import sklearn  # install scikit-learn`. To ignore an import, add an
    in-line comment beginning with `# skip install`.

    Other than `interactive`, all keyword arguments are passed to pip, e.g.
    `autoinstall(user=True, upgrade=True, index_url='https://example.com')`.

    This function does not look for transitive dependencies. Only import
    statements in the module where this function is called are considered.
    """
    try:
        upgrade = kwargs["upgrade"]
    except KeyError:
        upgrade = False

    filename = inspect.stack()[1].filename
    deps = _get_deps(filename)
    if not interactive:
        targets = []
        for mod, pkgs in deps:
            if pkgs:
                targets.extend(pkgs)
            else:
                targets.append(mod)
        if not upgrade:
            orig_targets, targets, installed_tgts = targets, [], []
            for t in orig_targets:
                if _is_installed(t):
                    installed_tgts.append(t)
                else:
                    targets.append(t)
            print(f'{" ".join(installed_tgts)} already installed.', file=sys.stderr)
        print(f'autoinstall is installing {" ".join(targets)} ...', file=sys.stderr)
        install(*deps, **kwargs)
    else:
        for mod, pkgs in deps:
            if pkgs:
                targets = pkgs
            else:
                targets = [mod]

            if not upgrade and all(_is_installed(pkg) for pkg in targets):
                print(f"{' '.join(targets)} already installed.", file=sys.stderr)
                continue

            while True:
                resp = input(
                    f"Install {' '.join(targets)}? (Y)es / (N)o / (C)ustomize name  "
                )  # noqa: E501
                if re.search(r"y(?:es)?", resp, flags=re.I):
                    print(f"Installing {' '.join(targets)}...", file=sys.stderr)
                    install(*targets, **kwargs)
                    break
                elif re.search(r"no?", resp, flags=re.I):
                    break
                elif re.search(r"c(?:ustom)?", resp, flags=re.I):
                    pkg_name = input(
                        f"What package(s) would you like to install instead of {' '.join(targets)}? "  # noqa: E501
                    )
                    print(f"Installing {pkg_name}...", file=sys.stderr)
                    install(*pkg_name.split(), **kwargs)
                    break
                else:
                    print(
                        'Invalid input. Type "y", "n", or "c" and press [enter].',
                        file=sys.stderr,
                    )



from pipster._experimental.autoinstall import _get_deps


def test_get_deps():
    assert _get_deps("tests/data/imports.py") == [
        ("bs4", ("beautifulsoup4",)),
        ("libcst", None),
        ("sklearn", ("scikit-learn",)),
        ("requests", None),
        ("numpy", None),
        ("pandas", None),
    ]



# from importlib import import_module
import re

from pipster.install import _build_install_cmd
from pipster.install import _get_dist_name
from pipster.install import _get_requirements_from_file
from pipster.install import _install

TEST_PKG = "realpython-reader"
TEST_PKG_A = TEST_PKG + "==0.0.1"
TEST_PKG_B = TEST_PKG + "==1.1.1"
TEST_MODULE = "reader"

WHEEL_PATH = "tests/data/packages/"
WHEEL1 = WHEEL_PATH + "simplewheel-1.0-py2.py3-none-any.whl"
WHEEL2 = WHEEL_PATH + "simplewheel-2.0-py2.py3-none-any.whl"

REQ_FILE = "tests/data/requirements.txt"


"""
TODO implement the following tests

_install(wheel1, target="/tmp//")
_install(wheel2, target="/tmp//", upgrade=True)
"""


def test_build_install_cmd_plain():
    io_pairs = [
        ("pip install some_pkg", ["pip", "install", "some_pkg"]),
        ("some_pkg", ["pip", "install", "some_pkg"]),
        (
            "pip install --user some_pkg",
            ["pip", "install", "--user", "some_pkg"],
        ),  # noqa: E501
        (
            "pip install -r requirements.txt",
            ["pip", "install", "-r", "requirements.txt"],
        ),  # noqa: E501
        (
            "pip install --target /tmp//dir_without_spaces /path/to/wheel",
            [
                "pip",
                "install",
                "--target",
                "/tmp//dir_without_spaces",
                "/path/to/wheel",
            ],
        ),  # noqa: E501
        (
            "pip install --target '/tmp//dir with spaces' /path/to/wheel",
            ["pip", "install", "--target", "/tmp//dir with spaces", "/path/to/wheel"],
        ),  # noqa: E501
    ]
    for arg, output in io_pairs:
        cmd = _build_install_cmd(arg)
        assert cmd == output


def test_build_install_cmd_kwarg():
    row_re = r"\|\s*`(.*?)`\s*\|\s*`(.*?)`\s*\|\s*$"
    with open("cli_options.md") as f:
        table_lines = [line for line in f if line.startswith("|")]
    option_lines = [line for line in table_lines if re.match(row_re, line)]
    assert len(option_lines) == len(table_lines) - 2
    for line in option_lines:
        cli_opt, kwarg = re.match(row_re, line).groups()
        evaluated = eval(f"_build_install_cmd({kwarg})")
        assert " ".join(evaluated[2:]) == cli_opt


def test__get_dist_name():
    assert _get_dist_name(WHEEL1) == "simplewheel"
    assert _get_dist_name(WHEEL2) == "simplewheel"
    assert _get_dist_name(TEST_PKG_A) == "realpython-reader"
    assert _get_dist_name(TEST_PKG_B) == "realpython-reader"
    urls = [  # from https://pip.pypa.io/en/stable/topics/vcs-support/
        "git+ssh://git.example.com/MyProject#egg=MyProject",
        "git+file:///home/user/projects/MyProject#egg=MyProject",
        "git+https://git.example.com/MyProject#egg=MyProject",
        "git+https://git.example.com/MyProject.git@master#egg=MyProject",
        "git+https://git.example.com/MyProject.git@v1.0#egg=MyProject",
        "git+https://git.example.com/MyProject.git@da39a3ee5e6b4b0d3255bfef95601890afd80709#egg=MyProject",
        "git+https://git.example.com/MyProject.git@refs/pull/123/head#egg=MyProject",
        "hg+http://hg.myproject.org/MyProject#egg=MyProject",
        "hg+https://hg.myproject.org/MyProject#egg=MyProject",
        "hg+ssh://hg.myproject.org/MyProject#egg=MyProject",
        "hg+file:///home/user/projects/MyProject#egg=MyProject",
        "hg+http://hg.example.com/MyProject@da39a3ee5e6b#egg=MyProject",
        "hg+http://hg.example.com/MyProject@2019#egg=MyProject",
        "hg+http://hg.example.com/MyProject@v1.0#egg=MyProject",
        "hg+http://hg.example.com/MyProject@special_feature#egg=MyProject",
        "svn+https://svn.example.com/MyProject#egg=MyProject",
        "svn+ssh://svn.example.com/MyProject#egg=MyProject",
        "svn+ssh://user@svn.example.com/MyProject#egg=MyProject",
        "svn+http://svn.example.com/svn/MyProject/trunk@2019#egg=MyProject",
        "svn+http://svn.example.com/svn/MyProject/trunk@{20080101}#egg=MyProject",
        "bzr+http://bzr.example.com/MyProject/trunk#egg=MyProject",
        "bzr+sftp://user@example.com/MyProject/trunk#egg=MyProject",
        "bzr+ssh://user@example.com/MyProject/trunk#egg=MyProject",
        "bzr+ftp://user@example.com/MyProject/trunk#egg=MyProject",
        # "bzr+lp:MyProject#egg=MyProject",  # TODO
        "bzr+https://bzr.example.com/MyProject/trunk@2019#egg=MyProject",
        "bzr+http://bzr.example.com/MyProject/trunk@v1.0#egg=MyProject",
        "MyProject @ git+ssh://git.example.com/MyProject",
        "MyProject @ git+file:///home/user/projects/MyProject",
        "MyProject @ git+https://git.example.com/MyProject",
        "MyProject @ git+https://git.example.com/MyProject.git@master",
        "MyProject @ git+https://git.example.com/MyProject.git@v1.0",
        "MyProject @ git+https://git.example.com/MyProject.git@da39a3ee5e6b4b0d3255bfef95601890afd80709",
        "MyProject @ git+https://git.example.com/MyProject.git@refs/pull/123/head",
        "MyProject @ hg+http://hg.myproject.org/MyProject",
        "MyProject @ hg+https://hg.myproject.org/MyProject",
        "MyProject @ hg+ssh://hg.myproject.org/MyProject",
        "MyProject @ hg+file:///home/user/projects/MyProject",
        "MyProject @ hg+http://hg.example.com/MyProject@da39a3ee5e6b",
        "MyProject @ hg+http://hg.example.com/MyProject@2019",
        "MyProject @ hg+http://hg.example.com/MyProject@v1.0",
        "MyProject @ hg+http://hg.example.com/MyProject@special_feature",
        "MyProject @ svn+https://svn.example.com/MyProject",
        "MyProject @ svn+ssh://svn.example.com/MyProject",
        "MyProject @ svn+ssh://user@svn.example.com/MyProject",
        "MyProject @ svn+http://svn.example.com/svn/MyProject/trunk@2019",
        "MyProject @ svn+http://svn.example.com/svn/MyProject/trunk@{20080101}",
        "MyProject @ bzr+http://bzr.example.com/MyProject/trunk",
        "MyProject @ bzr+sftp://user@example.com/MyProject/trunk",
        "MyProject @ bzr+ssh://user@example.com/MyProject/trunk",
        "MyProject @ bzr+ftp://user@example.com/MyProject/trunk",
        # "MyProject @ bzr+lp:MyProject",  # TODO
        "MyProject @ bzr+https://bzr.example.com/MyProject/trunk@2019",
        "MyProject @ bzr+http://bzr.example.com/MyProject/trunk@v1.0",
        "vcs+protocol://repo_url/#egg=MyProject&subdirectory=pkg_dir",
        "vcs+protocol://repo_url/#subdirectory=pkg_dir&egg=MyProject",
        "vcs+protocol://repo_url/#subdirectory=pkg_dir&egg=MyProject&",
    ]
    for url in urls:
        assert _get_dist_name(url) == "MyProject"
    assert _get_dist_name("MyProject @ path/to/MyProject_directory") == "MyProject"
    assert _get_dist_name("MyProject @ /path/to/MyProject_directory") == "MyProject"


def test__get_requirements_from_file():
    reqs = _get_requirements_from_file(REQ_FILE)
    assert reqs == [
        "pytest",
        "pytest-cov",
        "beautifulsoup4",
        "docopt == 0.6.1",
        'requests [security] >= 2.8.1, == 2.8.* ; python_version < "2.7"',
        "urllib3 @ https://github.com/urllib3/urllib3/archive/refs/tags/1.26.8.zip",
        "MyProject~=4.0.1",
        "other-pkg",
        "MyProject>=4.0.0",
        "./downloads/numpy-1.9.2-cp34-none-win32.whl",
        "http://wxpython.org/Phoenix/snapshot-builds/wxPython_Phoenix-3.0.3.dev1820+49a8884-cp34-none-win_amd64.whl",
    ]
    dist_names = [_get_dist_name(r) for r in reqs]
    assert dist_names == [
        "pytest",
        "pytest-cov",
        "beautifulsoup4",
        "docopt",
        "requests",
        "urllib3",
        "MyProject",
        "other-pkg",
        "MyProject",
        "numpy",
        "wxpython-phoenix",
    ]


def test_build_install_cmd_underscores():
    # find_links -> --find-links
    assert _build_install_cmd("some_pkg", find_links="/local/dir/") == [
        "pip",
        "install",
        "--find-links",
        "/local/dir/",
        "some_pkg",
    ]


def test_install_simplewheel():
    result1 = _install(WHEEL1)
    assert result1.returncode == 0
    result2 = _install(WHEEL2)
    assert result2.returncode == 0


def test_already_loaded(capsys):
    result1 = _install(WHEEL1)
    assert result1.returncode == 0
    import simplewheel  # type: ignore  # noqa: F401

    result2 = _install(WHEEL2)
    captured = capsys.readouterr()
    assert result2.returncode == 0
    assert "WARNING" in captured.out


# def test_ptp_already_loaded_warning():
#     # requires internet connection
#     _install(TEST_PKG_A)
#     import_module(TEST_MODULE)
#     with pytest.warns(UserWarning):
#         result = _install(TEST_PKG_B)
#         assert result.returncode == 0
#
#
# def test_ptp_already_loaded_warning_upgradeTrue():
#     # requires internet connection
#     _install(TEST_PKG_A)
#     import_module(TEST_MODULE)
#     with pytest.warns(UserWarning):
#         result = _install(TEST_PKG_B, upgrade=True)
#         assert result.returncode == 0



from glob import glob
import os
from pathlib import Path
import re
import shlex
import subprocess
import sys
from typing import List
from typing import Optional
from urllib.parse import urlparse
from warnings import warn

try:
    from pip._internal.commands.install import InstallCommand
    from pip._internal.network.session import PipSession
    from pip._internal.req.req_file import parse_requirements
    from pip._vendor.distlib.database import DistributionPath
    from pip._vendor.packaging.requirements import InvalidRequirement
    from pip._vendor.packaging.requirements import Requirement
    from pip._vendor.packaging.utils import InvalidSdistFilename
    from pip._vendor.packaging.utils import InvalidWheelFilename
    from pip._vendor.packaging.utils import parse_wheel_filename
    from pip._vendor.packaging.utils import parse_sdist_filename
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        f"Please install pip for the current interpreter: {sys.executable}"
    )

__all__ = ["install"]


install_cmd = InstallCommand(name="dummy", summary="Provides parse_args.")


def _check_for_pipfiles(depth=3):  # `3` taken from pipenv.utils.pipfile.find_pipfile()
    pipfiles = []
    for i in range(depth):
        pipfiles.extend(glob((".." + os.sep) * i + "Pipfile"))
    if pipfiles:
        pipfiles = [os.path.abspath(f) for f in pipfiles]
        msg = (
            "Warning: the following Pipfiles will be bypassed by "
            f"pipster.install: ({' '.join(pipfiles)})"
        )
        warn(msg, stacklevel=2)


def _check_if_already_loaded(requirements):
    req_targets = [_get_dist_name(r) for r in requirements]
    dist_path = DistributionPath(include_egg=True)
    target_providers = [
        d.modules
        for t in req_targets
        for d in dist_path.get_distributions()  # installed
        if t == d.name
    ]
    target_providers = [y for x in target_providers for y in x]  # flatten
    target_origins = {}
    for t in target_providers:
        if t in sys.modules:
            module_spec = sys.modules[t].__spec__
            if module_spec is not None:
                target_origins[module_spec.origin] = t
    dist_path = DistributionPath(include_egg=True)
    dists = [dist_path.get_distribution(t) for t in req_targets]
    paths = set()
    for dist in dists:
        if dist is not None:
            for path, _, _ in dist.list_installed_files():
                paths.add(os.path.join(os.path.dirname(dist.path), path))
    already_loaded = paths.intersection(target_origins)
    already_loaded = {target_origins[origin] for origin in already_loaded}
    return already_loaded


def _get_dist_name(target: str) -> Optional[str]:
    """Parse install target down to the distribution name."""
    warn_msg = (
        f"If {target} was already imported, python must be restarted "
        "to import the newly installed version."
    )

    # parse `pip install "asdf @ ..."`
    at_match = re.match(r"([A-Za-z0-9.-_]+)\s+@\s+(.+)", target)
    if at_match:
        at_name, target = at_match.groups()
    else:
        at_name = None

    if _is_valid_url(target):
        parsed_url = urlparse(target)
        if egg_match := re.search(r"egg=([A-Za-z0-9-_.]+)", parsed_url.fragment):
            return egg_match.group(1)
        elif re.search(r"\.(?:whl|tar\.gz|zip)$", parsed_url.path):
            filename = Path(parsed_url.path).name
            try:
                return _get_dist_name_from_filename(filename)
            except (InvalidSdistFilename, InvalidWheelFilename):
                return at_name
        else:
            if at_name:
                return at_name
            else:
                warn(warn_msg, UserWarning)
                return None  # TODO  Can this be determined further?
    elif re.search(r"\.(?:whl|tar\.gz|zip)$", target):
        filename = Path(target).name
        try:
            return _get_dist_name_from_filename(filename)
        except (InvalidSdistFilename, InvalidWheelFilename):
            return at_name
    else:
        try:
            return Requirement(target).name
        except InvalidRequirement:
            # Probably a directory. TODO use `os.path.isdir` and `setup.py egg_info`?
            if at_name:
                return at_name
            else:
                warn(warn_msg, UserWarning)
                return None


def _get_dist_name_from_filename(filename):
    if filename.endswith(".whl"):
        name, _, _, _ = parse_wheel_filename(filename)
        return name
    elif filename.endswith(".tar.gz") or filename.endswith(".zip"):
        name, _ = parse_sdist_filename(filename)
        return name
    else:
        raise ValueError(f"{filename} does not end in whl, tar.gz, or zip.")


def _is_valid_url(url):
    try:
        parsed_url = urlparse(url)
        return parsed_url.scheme and (
            parsed_url.netloc or parsed_url.scheme.endswith("file")
        )
    except:  # noqa: E722
        return False


def _get_requirements_from_file(filename):
    """Parse requirements.txt file into individual requirements."""
    return [r.requirement for r in parse_requirements(filename, PipSession())]


def install(*args, **kwargs) -> None:
    """Wrap _install to hide CompletedProcess in REPL output."""
    result = _install(*args, **kwargs)  # noqa: F841


def _install(*args, **kwargs) -> subprocess.CompletedProcess:
    """Install packages into the current environment.

    Equivalent examples of command-line pip and pipster are grouped below.

    METHOD 1: Single argument is exactly the same as command line interface,
    beginning with 'pip install ...'

    $ pip install --user --upgrade some_pkg
    >>> install('pip install --user --upgrade some_pkg')

    METHOD 2: Arguments from command-line implementation split on spaces

    $ pip install some_pkg
    >>> install('some_pkg')

    $ pip install --user --upgrade some_pkg
    >>> install('--user', '--upgrade', 'some_pkg')
    >>> install('some_pkg', user=True, upgrade=True)

    If preferred, keyword-value arguments can also be used:

    $ pip install -r requirements.txt
    >>> install('-r', 'requirements.txt')
    >>> install(r='requirements.txt')

    $ pip install --no-index --find-links /local/dir/ some_pkg
    >>> install('--no-index', '--find-links', '/local/dir/', 'some_pkg')
    # Note the use of '_' in the following keyword example.
    >>> install('some_pkg', index=False, find_links='/local/dir/')
    """
    _check_for_pipfiles()
    cli_args = _build_install_cmd(*args, **kwargs)
    # use pip internals to isolate package names
    req_args, req_targets = install_cmd.parse_args(cli_args)
    if req_args.requirements:  # -r requirements.txt
        reqs_from_file = _get_requirements_from_file(req_args.requirements)
        req_targets.extend(reqs_from_file)
    assert req_targets[:2] == ["pip", "install"]
    already_loaded = _check_if_already_loaded(req_targets[2:])
    print("Running `", " ".join(cli_args), "`  ...", file=sys.stderr)
    cli_cmd = [sys.executable, "-m"] + cli_args
    result = subprocess.run(cli_cmd, check=True)
    if result.returncode == 0 and already_loaded:
        print(
            "\n\033[0;31mWARNING:\033[00m The following modules were already loaded. "
            "Restart python to see changes:\n"
            f"\033[0;32m{'os.linesep'.join(already_loaded)}\033[00m\n"
        )
    return result


def _build_install_cmd(*args, **kwargs) -> List[str]:
    if len(args) == 1 and args[0].startswith("pip install "):
        return shlex.split(args[0])
    else:
        cli_args = ["pip", "install"]
        # Keyword arguments are translated to CLI options
        for raw_k, v in kwargs.items():
            k = raw_k.replace("_", "-")  # Python identifiers -> CLI long names
            if len(k) == 1:
                option = "-" + k
            else:
                option = "--" + k

            if v is None:
                # None->omit
                continue
            elif isinstance(v, bool):
                # true->include, false->include negated
                if k.startswith("no-"):
                    # suggest `some-option=True` instead of
                    # `no-some-option=False`
                    sugg_k = raw_k[3:]
                    raise ValueError(
                        f"Rather than '{raw_k}={v!r}', try '{sugg_k}={not v!r}'"
                    )
                if not v:
                    option = option.replace("--", "--no-")
                cli_args.append(option)
            elif isinstance(v, str):
                if not v:
                    raise ValueError(f"Empty string passed as value: ({raw_k}={v!r}")
                cli_args.extend([option, v])
            # Handle additive args (indicated by int value, e.g. q=3 -> -qqq)
            elif isinstance(v, int):  # v=1 (i.e. v=True) already handled above
                if len(k) == 1:
                    cli_args.append(f"-{k * v}")
                else:
                    cli_args.extend([option] * v)
                continue
            # Handle repeatable key-value arguments
            elif isinstance(v, list) or isinstance(v, tuple):
                if not all(isinstance(s, str) for s in v):
                    raise ValueError(
                        f"Not all elements of list/tuple are strings ({raw_k}={v!r})"
                    )
                for each_val in v:
                    cli_args.extend([option, each_val])
            else:
                raise ValueError(
                    "Value must be of type bool/str/int/list/tuple/None, "
                    f"not {type(v)} (from {raw_k}={v!r})"
                )
        # Positional arguments are passed directly as CLI arguments
        cli_args += args
    return cli_args



from bs4 import BeautifulSoup as bs  # type: ignore  # install beautifulsoup4
import libcst as cst  # type: ignore
from sklearn.datasets import load_iris  # type: ignore  # install scikit-learn
import requests as r  # type: ignore
import numpy, pandas  # type: ignore  # noqa: E401

# reference imported modules for style and type checkers
bs
cst
load_iris
r
numpy
pandas



from contextlib import AbstractContextManager
import re
import sys


# TODO Compare this thing: https://github.com/mdipierro/autoinstaller/blob/master/autoinstaller.py


class install_if_not_found(AbstractContextManager):
    """Context manager to handle ModuleNotFoundErrors
    Adapted from contextlib.suppress
    """

    def __enter__(self):
        pass

    def __exit__(self, exctype, excinst, exctb):
        print("exctype:", exctype, isinstance(exctype, ModuleNotFoundError))
        print("excinst:", excinst, isinstance(excinst, ModuleNotFoundError))
        print("exctb:", exctb, isinstance(exctb, ModuleNotFoundError))
        if isinstance(excinst, FileNotFoundError):
            print("doing stuff")
            failed_module = re.search(r"'((?:\\'|[^'])+)'", excinst.msg).group(1)
            while True:
                resp = input(
                    f"Could not import {failed_module}. "
                    "Install? (Y)es / (n)o / (c)ustomize  "
                )
                if re.search(r"y(?:es)?", resp, flags=re.I):
                    print("Installing {failed_module}...", file=sys.stderr)
                    break
                elif re.search(r"no?", resp, flags=re.I):
                    print("Okay. Taking no action.")
                    break
                elif re.search(r"c(?:ustom)?", resp, flags=re.I):
                    install_name = input(
                        "What package name would you like to " "install? "
                    )
                    print(f"Installing {install_name}...", file=sys.stderr)
                    break
                else:
                    print('Invalid input. Type "y", "n", or "c" and press [enter].')
        return exctype is not None and isinstance(excinst, ModuleNotFoundError)


# @contextmanager
# def install_if_not_found():
#     try:
#         yield
#     except ModuleNotFoundError as e:
#         print(dir(e))
#         print(e.msg)
#         failed_module = re.search(r"'((?:\\'|[^'])+)'", e.msg).group(1)
#         while True:
#             resp = input(f'Could not import {failed_module}. '
#                          'Install? (Y)es / (n)o / (c)ustomize  ')
#             if re.search(r'y(?:es)?', resp, flags=re.I):
#                 print('Installing {failed_module}...', file=sys.stderr)
#                 break
#             elif re.search(r'no?', resp, flags=re.I):
#                 print('Okay. Taking no action.')
#                 break
#             elif re.search(r'c(?:ustom)?', resp, flags=re.I):
#                 install_name = input(f'What package name would you like to install? ')
#                 print(f'Installing {install_name}...', file=sys.stderr)
#                 break
#             else:
#                 print('Invalid input. Type "y", "n", or "c" and press [enter].')
#

with install_if_not_found():
    # TODO move this to a pytest test
    import asdfasdf1  # type: ignore  # noqa: F401
    import asdfasdf1  # noqa: F401,F811
    import asdfasdf1  # noqa: F401,F811
